/**
* Copyright (C) 2017 Elisha Riedlinger
*
* This software is  provided 'as-is', without any express  or implied  warranty. In no event will the
* authors be held liable for any damages arising from the use of this software.
* Permission  is granted  to anyone  to use  this software  for  any  purpose,  including  commercial
* applications, and to alter it and redistribute it freely, subject to the following restrictions:
*
*   1. The origin of this software must not be misrepresented; you must not claim that you  wrote the
*      original  software. If you use this  software  in a product, an  acknowledgment in the product
*      documentation would be appreciated but is not required.
*   2. Altered source versions must  be plainly  marked as such, and  must not be  misrepresented  as
*      being the original software.
*   3. This notice may not be removed or altered from any source distribution.
*/

#include "cfg.h"

// Declare varables
bool p_StopThreadFlag = false;
bool p_ThreadRunningFlag = false;
HANDLE p_hThread = NULL;
DWORD p_dwThreadID = 0;

// Writes all bytes in Config to memory
void WriteAllByteMemory()
{
	HANDLE hProcess = GetCurrentProcess();

	for (UINT x = 1; x <= Config.AddressPointerCount && x <= Config.BytesToWriteCount; x++)
	{
		if (Config.MemoryInfo[x].AddressPointer > 0 && Config.MemoryInfo[x].SizeOfBytes > 0)
		{
			// Get current memory
			byte* lpBuffer = new byte[Config.MemoryInfo[x].SizeOfBytes];
			void* AddressPointer = (void*)Config.MemoryInfo[x].AddressPointer;
			if (ReadProcessMemory(hProcess, AddressPointer, lpBuffer, Config.MemoryInfo[x].SizeOfBytes, NULL))
			{
				// Make memory writeable
				unsigned long oldProtect;
				if (VirtualProtect(AddressPointer, Config.MemoryInfo[x].SizeOfBytes, PAGE_EXECUTE_READWRITE, &oldProtect))
				{
					// Write to memory
					memcpy(AddressPointer, Config.MemoryInfo[x].Bytes, Config.MemoryInfo[x].SizeOfBytes);

					// Restore protection
					VirtualProtect(AddressPointer, Config.MemoryInfo[x].SizeOfBytes, oldProtect, &oldProtect);

					// Delete varable
					delete Config.MemoryInfo[x].Bytes;

					// Store new variable
					Config.MemoryInfo[x].Bytes = lpBuffer;
				}
				else
				{
					Compat::Log() << "Access Denied at memory address: 0x" << std::showbase << std::hex << Config.MemoryInfo[x].AddressPointer << std::dec << std::noshowbase;
				}
			}
			else
			{
				Compat::Log() << "Failed to read memory at address: 0x" << std::showbase << std::hex << Config.MemoryInfo[x].AddressPointer << std::dec << std::noshowbase;
			}
		}
	}
}

// Verify process bytes before writing memory
bool CheckVerificationMemory()
{
	HANDLE hProcess = GetCurrentProcess();

	// Check Verification details
	if (Config.VerifyMemoryInfo.SizeOfBytes > 0 && Config.VerifyMemoryInfo.AddressPointer > 0)
	{
		// Get current memory
		byte* lpBuffer = new byte[Config.VerifyMemoryInfo.SizeOfBytes];
		if (ReadProcessMemory(hProcess, (LPVOID)Config.VerifyMemoryInfo.AddressPointer, lpBuffer, Config.VerifyMemoryInfo.SizeOfBytes, NULL))
		{
			for (UINT x = 0; x < Config.VerifyMemoryInfo.SizeOfBytes; x++)
			{
				if (lpBuffer[x] != Config.VerifyMemoryInfo.Bytes[x])
				{
					// Check failed
					return false;
				}
			}

			// All checks pass return true
			return true;
		}
	}

	// Default to fail
	return false;
}

// Thread to undo memory write after ResetMemoryAfter time
static DWORD WINAPI WriteMemoryThreadFunc(LPVOID pvParam)
{
	UNREFERENCED_PARAMETER(pvParam);

	// Get thread handle
	p_hThread = GetCurrentThread();

	// Set thread flag to running
	p_ThreadRunningFlag = true;

	// Sleep for a while
	DWORD timer = 0;
	while (!p_StopThreadFlag && timer < Config.ResetMemoryAfter)
	{
		Sleep(120);
		timer += 120;
	};

	// Logging
	Compat::Log() << "Undoing memory write...";

	// Undo the memory write
	WriteAllByteMemory();

	// Return value
	return 0;
}

// Main sub for writing to the memory of the application
void WriteMemory()
{
	if (CheckVerificationMemory())
	{
		// Logging
		Compat::Log() << "Writing bytes to memory...";

		// Writing bytes to memory
		WriteAllByteMemory();

		// Starting thread to undo memory write after ResetMemoryAfter time
		if (Config.ResetMemoryAfter > 0) CreateThread(NULL, 0, WriteMemoryThreadFunc, NULL, 0, &p_dwThreadID);
	}
	else
	{
		if (Config.VerifyMemoryInfo.AddressPointer != 0)
		{
			Compat::Log() << "Verification for memory write failed";
		}
	}
}

// Is thread running
bool IsWriteMemoryThreadRunning()
{
	return p_ThreadRunningFlag && GetThreadId(p_hThread) == p_dwThreadID && p_dwThreadID != 0;
}

// Stop WriteMemory thread
void StopWriteMemoryThread()
{
		// Set flag to stop thread
		p_StopThreadFlag = true;

		// Wait for thread to exit
		if (IsWriteMemoryThreadRunning()) WaitForSingleObject(p_hThread, INFINITE);

		// Close handle
		CloseHandle(p_hThread);
}